/***************************************************************
 * Name:      debuglog.c
 * Purpose:   print funtions to logfile in RAM
 * Author:    Nicolas Benezan (nicolas@benezan.de)
 * Created:   2024-01-16
 * Copyright: Nicolas Benezan (benezan-electronics.de)
 * License:

bene 13-Feb-2024 first experimental version
bene 23-Feb-2024 + lock, PrintMem() bugfix

 **************************************************************/
 
/* Printf()-alike logging functions that write to a buffer in
'' RAM which can be sent to the serial port post mortem

To enable real time debugging with as little execution time impact as possible
only pointers and integer numbers are written to a circular buffer in RAM.
Printf() formatting is done when the buffer is finally sent to the serial
port with DumpBuffer().

Each Print*() function writes a record to the buffer

long    size   ' -1 to -6 means string + zero to five 32 bit values
long    timestamp
long    string ' pointer to printf() style formatting string
long    arg[]  ' arguments

or

long    size   ' >0 means string + size number of longs
long    timestamp
long    string ' pointer to formatting string
byte    mem[]  ' copy of memory to print

records are always written continously, e.g. no split wraparound happens at 
the end of the buffer. If the maximum record size of 1kB doesn't fit the 
record is written to buffer[0]. This wastes up to 1kB of buffer space
in the worst case but avoids some special cases and thus simplifies and
speeds up writing a lot.
*/

#include <stdint.h>
#include <stdio.h>
#include <string.h> // memcpy()

#define BufLen     8192 // longs
#define maxBytes   1016 // bytes
#define maxRecSize 1024
#define maxStrSize 1024

static int head = 0; // write index for FIFO ring buffer
static int tail = 0; // read index
static int used = 0;
static int time0 = 0; // offset CT vs. displayed time
static uint32_t buffer[BufLen + maxRecSize/4];
static int LogLock;

void Alloc (nl)
// allocates nl number of longs by discarding the oldest records if necessary
{
  while (_locktry (LogLock)==0) {};
  used+= nl;
  while (used >= BufLen)
  {
    nl= buffer[tail];
    if (nl < 0) nl= -nl + 1;
    else nl+= 2;
    used-= nl;
    tail= (tail + nl) & (BufLen - 1);
  }
  head&= (BufLen - 1);
}

void Print0 (char* str)
{
  Alloc (3);
  buffer[head++]= -1;
  buffer[head++]= _cnt ();
  buffer[head++]= (int)str;
  _lockrel (LogLock);
}

void Print1 (char* str, int a1)
{
  Alloc (4);
  buffer[head++]= -2;
  buffer[head++]= _cnt ();
  buffer[head++]= (int)str;
  buffer[head++]= a1;
  _lockrel (LogLock);
}

void Print2 (char* str, int a1, int a2)
{
  Alloc (5);
  buffer[head++]= -3;
  buffer[head++]= _cnt ();
  buffer[head++]= (int)str;
  buffer[head++]= a1;
  buffer[head++]= a2;
  _lockrel (LogLock);
}

void Print3 (char* str, int a1, int a2, int a3)
{
  Alloc (6);
  buffer[head++]= -4;
  buffer[head++]= _cnt ();
  buffer[head++]= (int)str;
  buffer[head++]= a1;
  buffer[head++]= a2;
  buffer[head++]= a3;
  _lockrel (LogLock);
}

void Print4 (char* str, int a1, int a2, int a3, int a4)
{
  Alloc (7);
  buffer[head++]= -5;
  buffer[head++]= _cnt ();
  buffer[head++]= (int)str;
  buffer[head++]= a1;
  buffer[head++]= a2;
  buffer[head++]= a3;
  buffer[head++]= a4;
  _lockrel (LogLock);
}

void Print5 (char* str, int a1, int a2, int a3, int a4, int a5)
{
  Alloc (8);
  buffer[head++]= -6;
  buffer[head++]= _cnt ();
  buffer[head++]= (int)str;
  buffer[head++]= a1;
  buffer[head++]= a2;
  buffer[head++]= a3;
  buffer[head++]= a4;
  buffer[head++]= a5;
  _lockrel (LogLock);
}

void PrintMem (char* str, void* ptr, int bytes)
{
  if (bytes > 0 && bytes < maxBytes)
  {
    int nl= (bytes + 3)>>2;
    Alloc (nl + 3);
    buffer[head++]= nl;
    buffer[head++]= _cnt ();
    buffer[head++]= (int)str;
    memcpy (&buffer[head], ptr, nl*4);
    head+= nl;
    _lockrel (LogLock);
  }
}

void ResetLogTimer ()
{
  time0= _cnt ();
  LogLock= _locknew ();
}

void DumpBuffer ()
{
  printf ("*** DebugLog Buffer Dump: ***\n");
  while (used > 0)
  {
    while (_locktry (LogLock)==0) {};
    int nl= buffer[tail];
    int t= buffer[tail+1] - time0;
    printf ("%06d.%02dus ", t / 200, t % 200 / 2);
    //printf ("%09d ", t*5);
    char* str= (char*)buffer[tail+2];
    if (nl < 0) 
    {
        switch (nl)
        {
        case -1: 
            printf (str);
            break;
        case -2: 
            printf (str, buffer[tail+3]);
            break;
        case -3: 
            printf (str, buffer[tail+3], buffer[tail+4]);
            break;
        case -4: 
            printf (str, buffer[tail+3], buffer[tail+4], buffer[tail+5]);
            break;
        case -5: 
            printf (str, buffer[tail+3], buffer[tail+4], buffer[tail+5], buffer[tail+6]);
            break;
        case -6: 
            printf (str, buffer[tail+3], buffer[tail+4], buffer[tail+5], buffer[tail+6], buffer[tail+7]);
            break;
        }
        nl= -nl + 2;
    }
    else if (nl > 0)
    {
        int adr= 0;
        int bytes= nl * 4;
        uint8_t* ptr= (uint8_t*)&buffer[tail+3];
        printf (str);
        for (int row= 0; row < (nl + 3) / 4; row++)
        {
            printf ("%04x:", adr);
            for (int col= 0; col < 16; col++)
            {
                printf (" %02x", *ptr++);
                if (--bytes <= 0) break;
            }
            adr+= 16;
            printf ("\n");
        }
        nl+= 3;
    }
    else return;
    used-= nl;
    tail= (tail + nl) & (BufLen - 1);
    _lockrel (LogLock);
  }
}
